// Eingaberoutinen fr C++ mit Syntax- und Bereichskontrolle
// (c) 1996 by Andreas Meile, HTL Brugg-Windisch

#include <stdio.h>

// Optionen auswerten
int werte_optionen_aus(const char *eingabe, const char *optionen) {
  int opt_stat = -1, // Gewhlte Option
  i, j,              // Hilfszhler fr die Optionenauswertung
  gl,                // Flag fr Gleichheit
  opt_nr = 0;        // Zhler fr die Option

  i = 0;
  while(opt_stat == -1 && optionen[i++] == '/') {
    j = 0;
    gl = 1;
    while(optionen[i] != '/' && optionen[i]) {
      if(gl)
        gl = (eingabe[j++] == optionen[i]);
      i++;
    }
    if(gl && !eingabe[j])
      opt_stat = opt_nr;
    opt_nr++;
  }
  return opt_stat;
}

// Einlesen einer ganzen Zahl
int lies_ganze_zahl(const char *eing_aufford, const char *alt_opts,
int *wert) {
  char zeile[80];    // Puffer fr die Eingabezeile
  int zw_wert,       // Wert fr am Schluss, falls nicht die Vorgabe benutzt
                     // worden ist
      fuehrend,      // Anzahl fhrende Fremdzeichen
      end_pos,       // Endposition
      mit_vorg;      // Flag, ob der Vorgabewert in <> angezeigt werden muss

  mit_vorg = werte_optionen_aus("", alt_opts) != -1;
  do {
    if(mit_vorg)
      printf("%s <%d>? ", eing_aufford, *wert);
    else
      printf("%s? ", eing_aufford);
    gets(zeile);
    // Zuerst schauen, ob der Benutzer von einer Option Gebrauch gemacht hat
    zw_wert = werte_optionen_aus(zeile, alt_opts);
    if(zw_wert != -1)
      return zw_wert;
    fuehrend = -1; // initialisieren, da scanf() diese sonst in
    end_pos = -1; // bestimmten Situationen unverndert lsst
    sscanf(zeile, "%*[^+-0123456789] %n %d %n", &fuehrend, &zw_wert,
    &end_pos);
    // Bedingungen fr eine fehlerfreie Eingabe
    if(fuehrend == 0 // Es drfen keine Fremdzeichen am Anfang vorkommen
    && (end_pos == -1 || !zeile[end_pos]) // Keine Fremdzeichen am Schluss
    && zeile[0]) { // Es muss "Substanz" in Form von Ziffern enthalten sein
      *wert = zw_wert;
      return -1;
    }
    printf("Ungltige Eingabe! Bitte wiederholen\n");
  } while(1);
}

// Einlesen einer positiven Zahl (kompatibel zur Ganzzahl-Einleseroutine)
int lies_positive_zahl(const char *eing_auff, const char *alt_op,
int *wert) {
  int b_opt,   // Zwischenspeicher fr Option
      zw_wert; // Speicher fr Zwischenwert, um bei fehlerhafter Eingabe, um
               // den Vorgabewert nicht zu verlieren

  do {
    zw_wert = *wert;
    int b_opt = lies_ganze_zahl(eing_auff, alt_op, &zw_wert);
    if(zw_wert >= 1) {
      *wert = zw_wert;
      return b_opt;
    }
    printf("Ungltige Eingabe! Nur positive Werte zulssig\n");
  } while(1);
}

// Einlesen einer ganzen Zahl in einem Intervall
int lies_interv_ganze_zahl (const char *eing_auff, const char *alt_op,
int w_min, int w_max, int *wert) {
  int b_opt;    // Zwischenspeicher fr Option
  int zw_wert;  // Speicher fr Zwischenwert, um bei fehlerhafter Eingabe
                // den Vorgabewert nicht zu verlieren

  do {
    zw_wert = *wert;
    b_opt = lies_ganze_zahl(eing_auff, alt_op, &zw_wert);
    if(zw_wert >= w_min && zw_wert <= w_max) {
      *wert = zw_wert;
      return b_opt;
    }
    printf("Ungltige Eingabe! Nur Werte zwischen %d und %d zulssig\n",
    w_min, w_max);
  } while(1);
}

// Einlesen einer Gleitkommazahl
int lies_fliesskommazahl (const char *eing_aufford, const char *alt_opts,
float *wert) {
  char zeile[80];    // Puffer fr die Eingabezeile
  int gew_opt,       // Gewhlte Option
      fuehrend,      // Anzahl fhrende Fremdzeichen
      end_pos,       // Endposition
      mit_vorg;      // Flag, ob der Vorgabewert in <> angezeigt werden muss
  float zw_wert;     // Zwischenspeicher fr eingegebenen Wert

  mit_vorg = werte_optionen_aus("", alt_opts) != -1;
  do {
    if(mit_vorg)
      printf("%s <%f>? ", eing_aufford, *wert);
    else
      printf("%s? ", eing_aufford);
    gets(zeile);
    // Zuerst schauen, ob der Benutzer von einer Option Gebrauch gemacht hat
    gew_opt = werte_optionen_aus(zeile, alt_opts);
    if(gew_opt != -1)
      return gew_opt;
    fuehrend = -1; // initialisieren, da scanf() diese sonst in
    end_pos = -1; // bestimmten Situationen unverndert lsst
    sscanf(zeile, "%*[^+-.0123456789] %n %f %n", &fuehrend, &zw_wert,
    &end_pos);
    // Bedingungen fr eine fehlerfreie Eingabe
    if(fuehrend == 0 // Es drfen keine Fremdzeichen am Anfang vorkommen
    && (end_pos == -1 || !zeile[end_pos]) // Keine Fremdzeichen am Schluss
    && zeile[0]) { // Es muss "Substanz" in Form von Ziffern enthalten sein
      *wert = zw_wert;
      return -1;
    }
    printf("Ungltige Eingabe! Bitte wiederholen\n");
  } while(1);
}

// Einlesen einer Gleitkommazahl in einem Intervall
int lies_interv_fliesskommazahl (const char *eing_auff, const char *alt_op,
float w_min, float w_max, float *wert) {
  int b_opt;     // Zwischenspeicher fr Option
  float zw_wert; // Speicher fr Zwischenwert, um bei fehlerhafter Eingabe
                 // den Vorgabewert nicht zu verlieren

  do {
    zw_wert = *wert;
    b_opt = lies_fliesskommazahl(eing_auff, alt_op, &zw_wert);
    if(zw_wert >= w_min && zw_wert <= w_max) {
      *wert = zw_wert;
      return b_opt;
    }
    printf("Ungltige Eingabe! Nur Werte zwischen %f und %f zulssig\n",
    w_min, w_max);
  } while(1);
}
