// Kombinationen-Errater, hier als Verfolger aufgebaut
// basierend auf KOMERRA2.BAS, neue Features:
// - 32-Bit: Streichtabelle bis 4 Mia. Bits mglich
// - optimierte Speichernutzung: 1 Tabelleneintrag entspricht 1 Bit
// - dem System knnen Hilfen mitgegeben werden
// - Statusabfrage mglich: Das System meldet, was noch mglich ist
// - intelligentere Suche fr einen Vorschlag: Auswahl mit maximaler Streich-
//   rate in der Tabelle (statistisch), falls Misserfolg
// (c) 1990, 1997 by Andreas Meile, CH-8242 Hofen SH

#include <stdio.h>
#include <math.h>    // fr Berechnung, ob gengend gross
#include <stdlib.h>  // fr Zufallsgenerator
#include <time.h>    // fr Zeitmessung
#include "eing_rou.h"

const float t_max=32.0 * log(2.0);

int a_el = 4;        // Anzahl Elemente
int a_zu = 6;        // Anzahl Zustnde pro Element
int w_moe;           // Flag, ob Wiederholungen mglich

void main() {
  char zeile[80];

  printf("\033[0m\033[2J\033[2;1HKombinationenerrater Plus\n"
         "(c) 1990, 1997 by Andreas Meile, CH-8242 Hofen SH\n\n");

  int j_n;

  {
    unsigned long dummy;
    srandom(time(&dummy));  // entspricht einem RANDOMIZE TIMER
  }

  do {
    float t;     // Anzahl Mglichkeiten
    do {
      printf("Festlegung der Ratedimensionen\n\n");
      lies_positive_zahl("Anzahl Zustnde pro Element", "/", &a_zu);
      lies_positive_zahl("Anzahl Elemente", "/", &a_el);
      if (a_el <= a_zu) {
        int opts;
        do {
          printf("Wiederholungen mglich (J/N) <J>? ");
          gets(zeile);
          opts = werte_optionen_aus(zeile, "/n/N/j/J/");
          w_moe = opts >= 2;
          if(opts == -1)
            printf("Nur Eingabe J, j, N, n oder Leereingabe zulssig!\n");
        } while(opts == -1);
      } else
        w_moe = 1;
      // berprfen, ob nicht zu gross
      if(w_moe)   // Fall, wenn Wiederholung mglich
        t = float(a_el) * log(a_zu); // logarithmisch, um
      else {                         // berlufe zu verhindern
        t = 0;
        float u=float(a_zu);
        for(int i=0; i<a_el; i++) {
          t += log(u);
          u -= 1.0;
        }
      }
      if(t > t_max)
        printf("Die eingegebenen Werte bersteigen meine Rechenkapazitt!\n");
    } while(t > t_max);

    // Anzahl Kombinationen noch ganzzahlig ermitteln
    unsigned long anz_komb = 1;
    if(w_moe)
      for(int j=0; j < a_el; j++)
        anz_komb *= (unsigned long)(a_zu);
    else {
      int k=a_zu;
      for(int j=0; j < a_el; j++) {
        anz_komb *= (unsigned long)(k);
        k--;
      }
    }
    unsigned long anz_uebrig = anz_komb, // Anzahl Kombinationen
                  anz_uebrig_alt;        // fr Schlussvergleich:
                                         // Anzahl brige verwalten

    // Speicher allozieren
    int anz_ind = (anz_komb >> 5) + 1;

    clock_t  t1, t2;               // fr Stoppuhr

    unsigned long *str_tab = new unsigned long[anz_ind],
                  *s_inkr =  new unsigned long[a_el * (a_el + 1) + 1];

    int *zaehl = new int[a_el],    // Hochzhler
        *akt_k = new int[a_el],    // Aktuelle Kombination
        *zm1 = new int[a_zu],      // Auszhlen der Anzahl Vorkommen in der
                                   // Eingabe (gleichzeitig Streichtabelle)
        *zm2 = new int[a_zu],      // dito fr Kombination
        *zaehl_i = new int[a_el],  // innerer Hochzhler
        vorschl_flag = 0;          // Flag, ob zuletzt ein Vorschlag gemacht

    for(register int i=0; i<anz_ind; i++)
      str_tab[i] = 0xFFFFFFFFLU;  // Alle Bits auf 1 setzen = kommt noch in
      // Frage

    int option;

    while(anz_uebrig > 0) {
      // Hauptmen ausgeben
      if(anz_uebrig > 1) {
        if(vorschl_flag) // zuletzt ein Vorschlag gemacht
          option = 4;
        else if(anz_uebrig > 1300)  // falls noch viele Mglichkeiten offen sind
          option = 2;   // dann einfaches Verfahren fr einen Vorschlag,
        else            // um Rechenzeit zu sparen
          option = 3;   // sonst auf das intelligente Verfahren zurckgreifen
        printf("\nTotal %lu noch in Frage kommende Lsungskombinationen\n\n"
               "Bitte Funktion whlen\n"
               "1) Statusbericht\n"
               "2) Vorschlag fr Spielzug, einfacher Entscheid\n"
               "3) Vorschlag fr Spielzug, intelligenter Entscheid"
               " (rechenintensiv!)\n"
               "4) Spielzug mit Angaben\n"
               "5) Die Lsung lautet nicht ...\n"
               "6) Das k-te Element lautet x\n"
               "7) Das k-te Element lautet nicht x\n"
               "8) x kommt genau k mal vor\n", anz_uebrig);
        if(w_moe)
          printf("9) x kommt hchstens k mal vor\n"
                 "10) x kommt mindestens k mal vor\n"
          "11) Wiederholung einzelner Elemente kommen hchstens k mal vor\n");
        putchar('\n');
        lies_interv_ganze_zahl("Bitte Option whlen", "/", 1, (w_moe ? 11 : 8),
        &option);
      } else
        option = 12;

      int k, x,                   // die beiden Hilfsangaben k und x
          anz_vork,               // Anzahl Vorkommende fr Punkt 3
          anz_vork_ang,           // Flag, ob berhaupt angegeben
          anz_ueb,                // Anzahl bereinstimmende
          anz_ueb_ang,            // Flag, ob berhaupt angegeben
          gueltig,                // Flag fr gltige Eingabe
          weiser_zug = 0,         // Flag fr weiser Zug
          anz_ausg,               // Zhler fr die Anzahl Ausgaben
          anz_spal;               // Anzahl Spalten am Bildschirm

      unsigned long min_q,        // minimale brigbleibrate
                    tt,           // Lnge der Liste, Wahl fr Element
                    tt2;          // Hochzhler

      anz_uebrig_alt = anz_uebrig;

      // Vorbereitungen in Abhngigkeit der Option
      switch(option) {
        case 1 :  // Statusbericht
          anz_ausg = 0;
          anz_spal = 80 / (a_el + 1);  // Anzahl Spalten
          putchar('\n');
          break;
        case 2 :  // Zufallsvorschlag (zufllig freier Eintrag suchen)
          tt = (unsigned long)(double(random()) *
          ldexp(double(anz_uebrig), -31)); // Position bestimmen
          tt2 = 0;
          vorschl_flag = 1;
          break;
        case 3 :  // intelligenter Zugvorschlag
          {
            int opts;
            printf("P = Vorschlagauswahloptimierung nur aufgrund der \"Platz"
            " richtig\"-Information\n"
            "F = Optimierung aufgrund beider Informationen \"Platz und Farbe"
            " richtig\" fllen\n");
            do {
              printf("Optimierung nach P oder F <F>? ");
              gets(zeile);
              opts = werte_optionen_aus(zeile, "/p/P/f/F/");
              anz_vork_ang = opts >= 2;
              if(opts == -1)
                printf("Nur Eingabe P, p, F, f oder Leereingabe zulssig!\n");
            } while(opts == -1);
          }
          for(int i=0; i < a_zu; i++)
            zm1[i] = 0;
          min_q = 0xFFFFFFFFUL;
          vorschl_flag = 1;
          printf("Bitte etwas Geduld!\n");
          break;
        case 4 :
        case 5 :
          do {
            printf("Bitte eine Buchstabenkombination eingeben ");
            if(vorschl_flag) {      // Vorgabe mglich
              putchar('<');
              for(int i=0; i<a_el; i++)
                putchar('A' + akt_k[i]);
              printf("> ");
            }
            printf(": ");
            gets(zeile);
            if(vorschl_flag && zeile[0] == '\0') { // Vorgabewert verwenden
              for(int i=0; i<a_el; i++)
                zeile[i] = 'A' + akt_k[i];
              zeile[a_el] = '\0';
            }
            gueltig = 1;
            for(int i=0; i<a_zu; i++)
              zm1[i] = 0;
            for(i=0; gueltig && i<a_el; i++) {
              gueltig = zeile[i] >= 'A' && zeile[i] < 'A' + a_zu;
              if(gueltig) {
                akt_k[i] = zeile[i] - 'A';
                gueltig = w_moe || !zm1[akt_k[i]];
                if(gueltig)
                  zm1[akt_k[i]]++;            // gleichzeitig mitzhlen
              }
            }
            if(gueltig)
              gueltig = zeile[a_el] == '\0';  // muss die richtige Lnge
            // auzfweisen
            if(!gueltig)
              printf("Eingabe muss aus %d Zeichen zwischen A und %c bestehen"
              " (Gross/Kleinschreibung\nbeachten)%s\n", a_el, ('A'-1) + a_zu,
              w_moe ? "" : ", wobei keine Wiederholungen erlaubt sind");
          } while(!gueltig);
          if(option == 4) {
            anz_vork_ang = lies_interv_ganze_zahl("Wieviele Buchstaben kom"
            "men vor oder k=keine Angabe", "/k/K", 0, a_el, &anz_vork) == -1;
            anz_ueb_ang = lies_interv_ganze_zahl("Wieviele Buchstaben stim"
            "men davon berein (k=keine Angabe)", "/k/K", 0, anz_vork_ang ?
            anz_vork : a_el, &anz_ueb) == -1;
          }
          break;
        case 6 :
        case 7 :
          do {
            printf("Welcher Buchstabe kommt %svor? ", option == 4 ? "" :
            "nicht ");
            gets(zeile);
            gueltig = zeile[0] >= 'A' && zeile[0] < 'A' + a_zu &&
            zeile[1] == '\0';
            if(!gueltig)
              printf("Eingabe muss ein Buchstabe zwischen A und %c sein\n",
              ('A' - 1) + a_zu);
          } while(!gueltig);
          x = zeile[0] - 'A';
          lies_interv_ganze_zahl("Welche Stelle betrifft dies in der"
          " Kombination", "", 1, a_el, &k);
          k--;
          break;
        case 8 :
        case 9 :
        case 10 :
          do {
            printf("Betroffener Buchstabe? ");
            gets(zeile);
            gueltig = zeile[0] >= 'A' && zeile[0] < 'A' + a_zu &&
            zeile[1] == '\0';
            if(!gueltig)
              printf("Eingabe muss ein Buchstabe zwischen A und %c sein\n",
              ('A' - 1) + a_zu);
          } while(!gueltig);
          x = zeile[0] - 'A';
          lies_interv_ganze_zahl("Wieviele Male kommt er vor", "", 0,
          w_moe ? a_el : 1, &k);
          break;
        case 11 :
          k = 1;
          lies_interv_ganze_zahl("Maximal mgliche Anzahl Wiederholungen"
          " eines Elementes", "/", 1, a_el, &k);
      }

      // Stoppuhr starten
      t1 = clock();

      // Hilfszhler lschen
      for(i=0; i<a_zu; i++)
        zm2[i] = 0;

      // Kombinationengeneratorschleife
      register int *z = zaehl,  // Hilfszeiger
                   e = 0;       // Aufrufebene

      register unsigned long *t = str_tab, // Hilfszeiger
                             tw = *t,      // Wert bereitstellen
                             h = 1;        // Bitwert

      rekursion_ein:
      for(*z = 0; *z < a_zu; (*z)++) {
        if(w_moe || !zm2[*z]) {
          e++;
          zm2[*z]++;
          if(e < a_el) {
            z++;
            goto rekursion_ein;
            rekursion_aus:
            z--;
          } else {
            // Verarbeitung
            if(tw & h) {
              int streich = 0;
              switch(option) {
                case 1 : // Statusbericht
                  anz_ausg++;
                  if(anz_ausg <= 200) { // Nicht mehr als 200 Kombinationen
                    // ausgeben, um ein unntig langes Scrollen zu vermeiden
                    for(int i=0; i<a_el; i++)
                      putchar('A' + zaehl[i]);
                    putchar(anz_ausg % anz_spal ? ' ' : '\n');
                  }
                  break;
                case 2 : // Vorschlag mit Zufall machen
                  if (tt2 == tt)
                    for(i=0; i<a_el; i++)
                      akt_k[i] = zaehl[i];
                  tt2++;
                  break;
                case 3 : // intelligenter Vorschlag machen
                  {
                    // innere Kombinationengeneratorschleife
                    register int *z_i = zaehl_i, // Hilfszeiger
                                 e_i = 0;        // Aufrufebene

                    register unsigned long *t_i = str_tab, // Hilfszeiger
                                           tw_i = *t_i, // Wert bereitstellen
                                           h_i = 1,     // Bitwert
                                           q = 0;       // Aufsummierwert

                    // Initialisierungen
                    for(register int i=(anz_vork_ang ? a_el : 0); i >= 0; i--)
                      for(register int j=a_el; j >= i; j--)
                        s_inkr[i * a_el + j] = 0;

                    rekursion_ein_i:
                    for(*z_i = 0; *z_i < a_zu; (*z_i)++) {
                      if(w_moe || !zm1[*z_i]) {
                        e_i++;
                        zm1[*z_i]++;
                        if(e_i < a_el) {
                          z_i++;
                          goto rekursion_ein_i;
                          rekursion_aus_i:
                          z_i--;
                        } else {
                          // innere Verarbeitung
                          if(tw_i & h_i) {
                            register int h1=0, h2;
                            for(register int j=0; j<a_el; j++)
                              if(zaehl_i[j] == zaehl[j])
                                h1++;
                            if(anz_vork_ang) { // falls Anzahl Vorkommende
                              h2 = 0;          // angegeben
                              for(j=0; j<a_zu; j++)
                                h2 += (zm1[j] < zm2[j] ? zm1[j] : zm2[j]);
                                // Anzahl Vorkommende = Summe der Minima
                                // Aufsummierung
                              q += (s_inkr[h1 * a_el + h2]++);
                            } else
                              q += (s_inkr[h1]++);
                          }
                          h_i <<= 1;     // nchstes Bit
                          if(!h_i) {
                            h_i = 1;     // wieder neuer Shift-Wert
                            t_i++;       // nchster Speicher
                            tw_i = *t_i; // nchster Wert
                          }
                        }
                        zm1[*z_i]--;
                        e_i--;
                      }
                    }
                    if(e_i > 0)
                      goto rekursion_aus_i;
                    // Auswerten
                    if(q < min_q) { // Vorschlag besser als alles bisherige
                      min_q = q;
                      tt = 0;
                    }
                    if(q == min_q) {
                      tt++;
                      if((unsigned long)(random()) <= 0x80000000ul / tt)
                        // entspricht IF RND < 1! / CSNG(tt%) THEN ...
                        for(i=0; i<a_el; i++)
                          akt_k[i] = zaehl[i]; // bertragen; die obige
                        // Zufallsformel sorgt dafr, dass smtliche gleich-
                        // wertigen Flle stochastisch gleich stark sind
                    }
                  }
                  break;
                case 4 :  // Zug wurde angegeben
                  {
                    register int h1=0;
                    if(anz_vork_ang) {  // falls Anzahl Vorkommende angegeben
                      for(register int j=0; j<a_zu; j++)
                        h1 += (zm1[j] < zm2[j] ? zm1[j] : zm2[j]);
                        // Anzahl Vorkommende = Summe der Minima
                      if(h1 != anz_vork)  // abweichend?
                        streich = 1; // wenn ja, dann durchstreichen!
                    }
                    h1=0;
                    for(register int j=0; j<a_el; j++)
                      h1 += (zaehl[j] == akt_k[j]);
                    if(anz_ueb_ang && h1 != anz_ueb)
                      streich = 1; // durchstreichen
                    if(h1 == a_el)    // Alles stimmt berein =>
                      weiser_zug = 1; // dann war es ein brauchbarer Vorschlag
                    // ansonsten htte man zum vornherein sagen knnen, dass
                    // er nicht mehr mglich ist
                  }
                  break;
                case 5 :  // Garantiert ausgeschlossene Kombination
                  {
                    for(register int i=0; zaehl[i] == akt_k[i] && i<a_el;
                      i++);
                    if(i == a_el)  // bis am Ende gleich => gefunden
                      streich = 1;
                  }
                  break;
                case 6 : // k-tes Element lautet x
                  if(zaehl[k] != x)
                    streich = 1;
                  break;
                case 7 : // k-tes Element lautet nicht x
                  if(zaehl[k] == x)
                    streich = 1;
                  break;
                case 8 : // x kommt genau k mal vor
                  if(zm2[x] != k)
                    streich = 1;
                  break;
                case 9 : // x kommt hchstens k mal vor
                  if(zm2[x] > k)
                    streich = 1;
                  break;
                case 10 : // x kommt mindestens k mal vor
                  if(zm2[x] < k)
                    streich = 1;
                  break;
                case 11 : // Wiederholungen hchstens k-fach mglich
                  {
                    register int h1 = 0;
                    for(register int j=0; j < a_zu; j++)
                      if(zm2[j] > h1)
                        h1 = zm2[j];
                    if(h1 > k)
                      streich = 1;
                  }
                  break;
                case 12 : // Spezial: Lsung ausgeben
                  {
                    printf("Lsung gefunden, sie lautet ");
                    for(register int j=0; j < a_el; j++)
                      putchar('A' + zaehl[j]);
                    putchar('\n');
                    streich = 1;
                  }
              }
              if(streich) {
                tw &= ~h;     // streichen
                anz_uebrig--; // ebenfalls reduzieren
              }
            }
            h <<= 1;   // nchstes Bit
            if(!h) {
              *t = tw;  // Wert wieder ablegen
              h = 1;    // wieder neuer Shift-Wert
              t++;      // nchster Speicher
              tw = *t;  // nchster Wert
            }
          }
          zm2[*z]--;
          e--;
        }
      }
      if(e > 0)
        goto rekursion_aus;
      *t = tw;          // letzter Wert wieder ablegen

      // Stoppuhr anhalten
      t2 = clock();

      switch(option) {
        case 1 : // Statusbericht
          if (anz_ausg != anz_uebrig) {   // Sollte nie vorkommen!
            printf("\nInterner Fehler\n");
            exit(1);
          }
          if(anz_ausg > 200)
            for(int i=0; i<a_el; i++)
              putchar('.');
          putchar('\n');
          break;
        case 2 : // Vorschlag fr eine Kombination
        case 3 : // dito, intelligent
          printf("\nGeeigneter Zugvorschlag : ");
          for(i=0; i<a_el; i++)
            putchar('A' + akt_k[i]);
          putchar('\n');
          break;
        case 4 : // Versuch mit Angaben
          if(!weiser_zug)
            printf("\nAnmerkung zu diesem Zug: Sehr schlecht gewhlt, denn die"
            " vorherigen Angaben\nreichten bereits fr eine Ausscheidung "
            "aus.\n");
      }
      if(option >= 4 && option <= 11) // Report
        if(anz_uebrig == anz_uebrig_alt)
          printf("\nDie vorherige Information war leider nutzlos.\n");
        else
          printf("\nReduktion der Mglichkeiten um %lu\n", anz_uebrig_alt -
          anz_uebrig);
      if(option >= 4)     // Vernderung hat stattgefunden
        vorschl_flag = 0;

      // Schritt-Rechenzeit ausgeben
      if(option >= 2 && option <= 11)
        printf("Bentigte Rechenzeit fr diesen Schritt: %.2f Sek.\n",
        float(t2 - t1) / float(CLOCKS_PER_SEC));
    }
    if(option != 12)
      printf("\nMit den obigen Angaben lsst sich keine Lsung finden, "
      "es\nsteckt hchstwahrscheinlich Betrug dahinter.\n");

    // Speicher wieder freigeben
    delete str_tab;
    delete s_inkr;
    delete zaehl;
    delete akt_k;
    delete zm1;
    delete zm2;
    delete zaehl_i;

    do {
      printf("\nNoch einmal (J/N)? ");
      gets(zeile);
      j_n = werte_optionen_aus(zeile, "/n/N/j/J");
      if(j_n == -1)
        printf("Nur Eingabe J, j, N, n oder Leereingabe zulssig!\n");
    } while(j_n == -1);
  } while(j_n >= 2);
}
